function [PRNU] = generateReferencePattern(dirPath,extension,uniqueField)
%% HELP
% Generates a refrence pattern from a set of images that have all been
% taken from a single known source.
%
% Parameter entries include a folder with the images contained there in,
% the extension you wish to examine (images must all be of the same type or
% fingerprint estimations will fail) and a uniqueField ID that will be used
% to save the Refrence Pattern at the completion of the program run.
%
% LIMITATIONS
% --> Current iteration of method is specifically encoded to work with
%     Pentax *istDL .PEF images only.
%
% Last Update: 13th of May 2015
% Coded by Richard Matthews 
% The University of Adelaide
%
% FUTURE WORK
% --> Extract Make, Model etc to create a more meaningful savefile 
%       --> can do this via EXIF iff EXIF data can be trusted.
% --> Extract image size data from file prior to processing
% --> Conduct checks to make sure only files of same size are used
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% PRELIM

cd(dirPath);
contents = dir([dirPath,extension]);
saveFile = uniqueField; %This value is used to identify the RP when it is saved.

% read in from EXIF data the size of the images that are being tested
% currently this is assumed for pentax at 2024x3040
% before method overloading occurs this will need to be modified to ensure
% additional image sizes can be handled without the magic numbers being
% encoded.


% Image File Types Supported  %% ** ONLY PEF WORKS CURRENTLY ** %%
File_TIF = '.tif';
File_TIFF= '.tiff';
File_PEF = '.PEF';
File_BMP = '.bmp'; 
File_jpg = '.jpg'; 
File_JPG = '.JPG';
File_CANNON = '.CR2'; %#ok<NASGU>
noSave = 'noSave';

database_size = length(contents);       %Number of Images
%sizeof = [n,m];

%if(n<1500 || m<1500)
%    error('Images are smaller than 1500x1500 pixels. Not suitable for reference pattern approximation. Exiting.');
%end
sizeof = [1024,1024]; %Hard limit to 150 pixels

%% setup PRNU array
blockSize=[256,256]; 
advance=[128,128];
numberOfCalls = 0.5*sizeof;
numberOfCalls(1,1)=(numberOfCalls(1,1)./blockSize(1,1))+(numberOfCalls(1,1)./blockSize(1,1)-1);
numberOfCalls(1,2)=(numberOfCalls(1,2)./blockSize(1,2))+(numberOfCalls(1,2)./blockSize(1,2)-1);
%matrix = ceil([sizeof./smith].*(2.*increase)./sizeof) + ceil(sizeof./smith);


% Check folder for number of images
images=length(contents);

%Make PRNU
%PRNU =zeros(sizeof(1)-matrix(1,1).*20,sizeof(2)-matrix(1,2).*20,3,images);
%PRNU = zeros(1454,1454,3,images); %HARDCODED FOR RASPI WORK THIS OUT LATER: WILL WORK FOR 1500x1500 crop.
%PRNU = double(PRNU); 
PRNU = zeros(2.*(numberOfCalls(1,1)*(blockSize(1,1)-advance(1,1))),2.*(numberOfCalls(1,1)*(blockSize(1,1)-advance(1,1))),3,images);
PRNU = double(PRNU);
%whos PRNU
%% error check

if database_size ==0,
    error('No images of specified type in the directory');
end

%% routine start
%count = 0;
%replace parfor
parfor j=1:1:database_size
    filename=contents(j).name;          % get filename with extension to parse to lower methods
    [~,fileName,fileExtension]=fileparts(contents(j).name);    % get fileName and fileExtension to conduct tests upon
    
    %count=count+1;

    %fprintf('\nProcessing image: ');
    %fprintf(fileName);
    %fprintf(fileExtension);
    %fprintf('\nImage Count: %10.0f', count');
    %fprintf('\n');

    %% Read in the image
    %fprintf('\nImage Read:....');
    % Pentax Image format
    if strcmp(fileExtension,File_PEF)            
        [picture] = readPentaxImage(filename);  % 12bit signal
        picture = picture(38:1061,38:1061,:); %crop image to 1024 x 1024 pixels
        %picture = double255(picture);% scale to 0-255
        %fprintf('\t\t\t\t\tPentax');
        picture = int16(picture);
        
    elseif strcmp(fileExtension,File_TIF)            
        [picture] = readTiffImage(filename);  % 12bit signal
        picture = picture(38:1061,38:1061,:); %crop image to 1024 x 1024 pixels
        %picture = double255(picture)% scale to 0-255
        %fprintf('\t\t\t\t\tTif');
    elseif strcmp(fileExtension,File_TIFF)
        [picture] = readTiffImage(filename);
        picture = picture(38:1061,38:1061,:); %crop image to 1024 x 1024 pixels
        %picture = double255(picture);% scale to 0-255
        %picture = picture.*64; %shift left 6
        picture=int16(picture);
        %whos picture
        %fprintf('\t\t\t\t\tTiff');
    elseif strcmp(fileExtension,File_BMP)
        continue
    elseif strcmp(fileExtension,File_JPG)
        [picture] = imread(filename); %8-bit depth
       picture = picture(38:1061,38:1061,:); %crop image to 1024 x 1024 pixels
        %picture = double255(picture);% scale to 0-255
       % fprintf('\t\t\t\t\tJPEG');
    elseif strcmp(fileExtension,File_jpg)
        [picture] = imread(filename); %8-bit depth
        %picture = single(picture);
        picture = picture(38:1061,38:1061,:); %crop image to 1024 x 1024 pixels
        %picture = double255(picture);% scale to 0-255
        %fprintf('\t\t\t\t\tJPEG');
    else
        %fprintf('\nImage type not supported.\n')
        continue
    end
%% SCALE IMAGE
%%picture = 1.3*picture; %issue with pixels under value 10.
%picture = double(picture);


    
    %% Removal of Saturated pixels 
    
    % for a 12-bit signal only pixels in the range of 0-3935 are valid
    % these pixels will be replaced with values of 0
    % 10 bit = 1024 therefore shitf
    
    if strcmp(fileExtension,File_PEF)
        picture(picture>(2^12-0.05*2^12))=0; %we define a hot pixel as one with a value of 95% or higher
    elseif strcmp(fileExtension,File_TIF)
        picture(picture>(2^16-0.05*2^16))=0;
    elseif strcmp(fileExtension,File_TIFF)
        picture(picture>(2^10-0.05*2^10))=0; 
        %picture = picture./64; %shift from 16bit down to 10 
    elseif strcmp(fileExtension,File_BMP)
        continue
    elseif strcmp(fileExtension,File_JPG)
         %ignore for now... impossible to ascertain which pixels were saturated.
    elseif strcmp(fileExtension,File_jpg)
         %ignore for now... impossible to ascertain which pixels were saturated.
    else
        fprintf('\nImage type not supported.\n')
        continue
    end
    
   %whos picture
   %
 %% normalise signal
 %   picture = (picture)./(2^10-1); 
    

    %% Layerise the image

            CFA = getCameraCFA(filename);
            if strcmp(fileExtension,File_PEF)   
                %fprintf('\nDemosaic Image:....');
                [picture, ~, ~, ~] = Demon(picture,CFA);  % 12bit signal
                %fprintf('\t\t\t\tSuccess');
                %picture = double255(layers);
            elseif strcmp(fileExtension,File_TIF)         
                %fprintf('\nDemosaic Image:....');
                [picture, ~, ~, ~] = Demon(picture,CFA);  % 10bit signal
                %fprintf('\t\t\t\tSuccess');
                %picture = double255(layers);
            elseif strcmp(fileExtension,File_TIFF)
                %fprintf('\nDemosaic Image:....');
                [picture, ~, ~, ~] = Demon(picture,CFA);  % 10bit signal
                %fprintf('\t\t\t\tSuccess');
                %picture = double255(layers);
             elseif strcmp(fileExtension,File_BMP)
                continue
             elseif strcmp(fileExtension,File_JPG)

            elseif strcmp(fileExtension,File_jpg)    

            else
                fprintf('\nImage type not supported.\n')
                continue
            end
%whos picture

    %% get noise estimations per layer
       % disp(j);
       %whos PRNU
       PRNU(:,:,:,j) = single(getNoiseEstimation(picture,fileName,CFA));
       %PRNU(:,:,:,j) = single(NoiseExtract(X(:,:,j),qmf,2,L));
       %whos PRNU
        
          
end
%whos PRNU

%% Residuals

            % four options for each location of CFA1 CF2 CF3 and CF4
            % for CFA1 array remove even rows and even col
            CFA1 = picture(1:2:end,1:2:end,1);

            % for CFA2 array remove even rows and odd col
            CFA2 = picture(1:2:end,2:2:end,2);
 
            % for CFA3 array remove odd rows and even col
            CFA3 = picture(2:2:end,1:2:end,2);

            % for CFA4 array remove odd rows and odd col
            CFA4 = picture(2:2:end,2:2:end,3);
blockSize=[256,256]; 
advance=[128,128];

% Test image arrays for length issues
[numberOfCalls(1,1),numberOfCalls(1,2)] = size(CFA1(:,:,1));
numberOfCalls(1,1)=(numberOfCalls(1,1)./blockSize(1,1))+(numberOfCalls(1,1)./blockSize(1,1)-1);
numberOfCalls(1,2)=(numberOfCalls(1,2)./blockSize(1,2))+(numberOfCalls(1,2)./blockSize(1,2)-1);
            
shiftCFA1    =circshift(CFA1,-0.5.*advance(1,1),1);
shiftCFA1    =circshift(shiftCFA1,-0.5.*advance(1,2),2);

shiftCFA2    =circshift(CFA2,-0.5.*advance(1,1),1);
shiftCFA2    =circshift(shiftCFA2,-0.5.*advance(1,2),2);

shiftCFA3    =circshift(CFA3,-0.5.*advance(1,1),1);
shiftCFA3    =circshift(shiftCFA3,-0.5.*advance(1,2),2);

shiftCFA4    =circshift(CFA4,-0.5.*advance(1,1),1);
shiftCFA4    =circshift(shiftCFA4,-0.5.*advance(1,2),2);


% [a,b,c,d] = size(patternNoiseCFA1);
%remove end elements
% trimmedCFA1 = zeros(a-advance(1,1),b-advance(1,2),c,d);
% trimmedCFA2 = trimmedCFA1;
% trimmedCFA3 = trimmedCFA1;
% trimmedCFA4 = trimmedCFA1;

%trimmedCFA1 = shiftCFA1(1:1:end-advance(1,1),1:1:end-advance(1,2));
%trimmedCFA2 = shiftCFA2(1:1:end-advance(1,1),1:1:end-advance(1,2));
%trimmedCFA3 = shiftCFA3(1:1:end-advance(1,1),1:1:end-advance(1,2));
%trimmedCFA4 = shiftCFA4(1:1:end-advance(1,1),1:1:end-advance(1,2));

% Remove Image DC

%DCRemovedCFA1= trimmedCFA1 - mean(mean(trimmedCFA1));
%DCRemovedCFA2= trimmedCFA2 - mean(mean(trimmedCFA2));
%DCRemovedCFA3= trimmedCFA3 - mean(mean(trimmedCFA3));
%DCRemovedCFA4= trimmedCFA4 - mean(mean(trimmedCFA4));

%FFT method
% FTCFA1 = fft2(trimmedCFA1);
% FTCFA1(1,1) = 0;
% DCRemovedCFA1=ifft2(FTCFA1);
% 
% FTCFA2 = fft2(trimmedCFA2);
% FTCFA2(1,1) = 0;
% DCRemovedCFA2=ifft2(FTCFA2);
% %fprintf('\nAssigning Blocks:....');

%matrix now contains the number of blocks in mxn 
% FTCFA3 = fft2(trimmedCFA3);
% FTCFA3(1,1) = 0;
% DCRemovedCFA3=ifft2(FTCFA3);
% 
% FTCFA4 = fft2(trimmedCFA4);
% FTCFA4(1,1) = 0;
% DCRemovedCFA4=ifft2(FTCFA4);

Red = zeros(2.*(numberOfCalls(1,1)*(blockSize(1,1)-advance(1,1))),2.*(numberOfCalls(1,1)*(blockSize(1,1)-advance(1,1))));
Green = Red;
Blue = Red;

% create sparse image in typical file format
%Red(1:2:end,1:2:end) = DCRemovedCFA1;
%Green(1:2:end,2:2:end) = DCRemovedCFA2;
%Green(2:2:end,1:2:end) = DCRemovedCFA3;
%Blue(2:2:end,2:2:end) = DCRemovedCFA4;

Red(1:2:end,1:2:end)= trimmedCFA1;
Green(1:2:end,2:2:end) = trimmedCFA2;
Green(2:2:end,1:2:end) = trimmedCFA3;
Blue(2:2:end,2:2:end) = trimmedCFA4;
% perform k = Y-I
PICTURE = cat(3,Red,Green,Blue);


P = PICTURE - PRNU;


%% Find Average of PRNU
PRNU = mean(PRNU,4);
%fprintf('\nNoise Approximation Saved.');
%fprintf('\n');

%% combine layers
%GreyScale = (0.3.*PRNU(:,:,1,1)+ 0.6.*PRNU(:,:,2,1) + 0.1.*PRNU(:,:,3,1));

%% publish Refrence Pattern
%ReferencePattern = PRNU;
%ReferencePattern = int16(PRNU);% return full 12-bit fingerprint
%ReferencePattern_8bit = ReferencePattern./4095.*255; % truncated 8-bit fingerprint

%maximum = max(max(ReferencePattern));
%minimum = min(min(ReferencePattern));
%average = mean(mean(ReferencePattern));

%fprintf('\n');

%% ERROR CHECK
if(images<50)
    fprintf('\nWARNING: less than 50 images analysed.');
    fprintf('\n         reference pattern may not be accurate.');
    fprintf('\n');
end

if (strcmp(saveFile,noSave))
else
    save(saveFile,'PRNU'); 
end


%%% FUNCTIONS %%%
function X = double255(X)
%convert to double ranging from 0 to 255
datatype = class(X);
    switch datatype,            % convert to [0,255]
        case 'uint8',  X=double(X);
        case 'uint16', X=ceil(double(X)/4096*255);
    end

